caption {
    padding: 0px;
    caption-side: top;
    font-weight: bold;
    color: rgba(var(--pst-color-paragraph),1);
}

/* Makes the image on each card in the intro smaller. */
.intro-card .card-img-top {
    max-width: 50%;
    padding: 20px;
    margin: auto;
}

/* Removes some of the empty space at the bottom of each intro card. */
.intro-card .card-body {
    padding-bottom: 0;
}
.intro-card .card-body .card-text {
    margin-bottom: 0;
}

/* Reduces the space around tab labels. */
.tabbed-set > label {
    padding: 0.5em 0.8em;
}

/* Format for page banners. Based on the admonition styling. */
.tmlt-banner-warning {
  margin: 0 auto 1.5625em;
  padding: .4rem .8rem;
  overflow: hidden;
  border-radius: .2rem;
  /* from --pst-color-warning */
  background-color: rgba(238, 144, 64, .2);
  border-left: .2rem solid var(--pst-color-warning);
  box-shadow: 0 .2rem .5rem var(--pst-color-shadow), 0 0 .0625rem var(--pst-color-shadow);
}
